import pickle
import hmac
import hashlib


class TrustedPickle:
    def __init__(self, secret_key):
        self.secret_key = secret_key

    def secure_write(self, model_path, model):
        """
            Attaches a secret_key to the model object and writes it to a file.
            Parameters:
                model_path (path-to-save-model-file): file location
                model (object referring to the machine learning model): python object
            Returns:
                file: secure model file.
        """

        # Convert the model object to a bytes object instead of writing to a file.
        data = pickle.dumps(model)

        # Calculate the signature and use digest to generate the byte stream.
        signature = hmac.new(self.secret_key, data, hashlib.sha256).digest()

        # Write the secured + serialized model to file
        with open(model_path, "wb") as f:
            f.write(signature + data)

    def verify_signature(self, untrusted_data):
        """
            Verifies the signature by reading back the secret_key, if the secret_key matches, then the signature is
            verified. Otherwise, it fails.
            Parameters:
                untrusted_data (data read from the file): string
            Returns:
                boolean: true if verified, raises an exception otherwise.
        """
        # Calculate the hashed signature to perform verification
        signature_length = len(hashlib.sha256(self.secret_key).digest())
        # Adding a try/except block to handle exceptions while verifying signature.
        try:
            if len(untrusted_data) >= signature_length:
                read_signature = untrusted_data[:signature_length]
                data = untrusted_data[signature_length:]
                computed_signature = hmac.new(self.secret_key, data, hashlib.sha256).digest()
                assert hmac.compare_digest(computed_signature, read_signature)
                return data
        except AssertionError:
            print("Warning!! Data has been tampered!!")
            # EDITED TO ADD:
            raise Exception  # Note: if needed, we can define a custom SecurityError exception type, or something

    def secure_read(self, file_path):
        """
           Reads the model from file and verifies the digital signature attached to it.
           Parameters:
                file_path (path-to-saved-model-file): file location
           Returns:
                trusted_data: model object (having verified the digital signature).
        """
        with open(file_path, "rb") as f:
            data_to_unpickle = f.read()
        read_data = self.verify_signature(data_to_unpickle)
        trusted_data = pickle.loads(read_data)
        return trusted_data
